"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const amqplib_1 = require("amqplib");
const ReadyGate_1 = require("../util/ReadyGate");
const RabbitmqConfig_1 = require("./RabbitmqConfig");
class RabbitmqClient {
    constructor(clientConfig, name) {
        /**
         * Whether the connection should be closed. This is not a statud indicator that show whether the connection is closed.
         * It is a flag that indicates whether the client has been purposely closed by code, as opposed to being closed because of an error.
         */
        this.closed = false;
        this.reconnecting = false;
        this.readyGate = new ReadyGate_1.ReadyGate();
        this.connectFunction = amqplib_1.connect;
        this.clientConfig = Object.assign({ protocol: 'amqp', maxConnectionAttempts: RabbitmqConfig_1.DEFAULT_MAX_CONNECTION_ATTEMPTS }, clientConfig);
        this.name = name;
        this.readyGate.channelNotReady();
    }
    async init() {
        await this.connect();
    }
    /**
     * Connect to RabbitMQ broker
     */
    async connect() {
        await this.createConnection();
        await this.createChannel();
        this.reconnecting = false;
        this.readyGate.channelReady();
    }
    async createConnection() {
        const newConnection = await this.connectFunction(this.clientConfig);
        newConnection.on('close', () => { this.handleErrorOrClose('Connection Closed'); });
        newConnection.on('error', (err) => { this.handleErrorOrClose('Connection Error', err); });
        this.connection = newConnection;
        this.logger.info('Connection established');
    }
    async createChannel() {
        const newChannel = await this.connection.createChannel();
        newChannel.on('close', (err) => { this.handleErrorOrClose('Channel closed unexpectedly', err); });
        newChannel.on('error', (err) => { this.handleErrorOrClose('Channel error', err); });
        this.channel = newChannel;
        this.logger.info('Channel opened');
    }
    handleErrorOrClose(cause, err) {
        if (err) {
            this.logger.error(err, `${cause}. Reconnecting`);
        }
        else {
            this.logger.error(`${cause}. Reconnecting`);
        }
        this.closeAllAndScheduleReconnection();
    }
    async closeAllAndScheduleReconnection() {
        if (!this.reconnecting) {
            this.readyGate.channelNotReady();
            this.reconnecting = true;
            this.logger.warn('Initialising reconnection...');
            if (this.channel) {
                this.logger.debug('Closing channel');
                try {
                    await this.closeChannel();
                }
                catch (e) {
                    // Do nothing... we tried to play nice
                }
            }
            if (this.connection) {
                this.logger.debug('Closing connection');
                try {
                    await this.closeConnection();
                }
                catch (e) {
                    // Do nothing... we tried to play nice
                }
            }
            this.reconnectionTimer = setTimeout(() => this.attemptReconnection(), 500);
        }
    }
    backoffWait(tryNum) {
        const waitBase = Math.min(Math.pow(3, Math.max(0, tryNum - 1)), 30) * 1000;
        // 1 second, 3 seconds, 9 seconds, 27 seconds, 30 seconds, 30 seconds, ....
        const waitMillis = waitBase + (Math.round(Math.random() * 800));
        return new Promise((resolve) => setTimeout(resolve, waitMillis));
    }
    async attemptReconnection() {
        this.clearReconnectionTimer();
        let attempts = 0;
        while (attempts < this.clientConfig.maxConnectionAttempts) {
            attempts++;
            await this.backoffWait(attempts);
            try {
                await this.connect();
                return;
            }
            catch (e) {
                this.logger.warn(e, `Failed reconnection attempt #${attempts}`);
            }
        }
        this.logger.error(`Couldn't reconnect after ${this.clientConfig.maxConnectionAttempts} attempts`);
        this.reconnecting = false;
        // tslint:disable-next-line
        if (this.clientConfig.exitOnIrrecoverableReconnect !== false) {
            this.logger.error('Cowardly refusing to continue. Calling shutdown function');
            this.shutdownFunction();
        }
    }
    clearReconnectionTimer() {
        if (this.reconnectionTimer) {
            clearTimeout(this.reconnectionTimer);
            this.reconnectionTimer = undefined;
        }
    }
    async close() {
        this.closed = true;
        await this.closeChannel();
        await this.closeConnection();
    }
    async closeChannel() {
        if (this.channel) {
            this.channel.removeAllListeners();
            this.channel.close();
            this.channel = undefined;
        }
    }
    async closeConnection() {
        if (this.connection) {
            this.connection.removeAllListeners();
            this.connection.close();
            this.connection = undefined;
        }
    }
}
exports.RabbitmqClient = RabbitmqClient;
//# sourceMappingURL=RabbitmqClient.js.map